import { __assign } from "tslib";
import { defaultLocale } from './locale';
import { startOfWeekYear } from './util';
var formattingTokens = /(\[[^\[]*\])|(MM?M?M?|Do|DD?|ddd?d?|w[o|w]?|YYYY|YY|a|A|hh?|HH?|mm?|ss?|S{1,3}|x|X|ZZ?|.)/g;
var match1 = /\d/; // 0 - 9
var match2 = /\d\d/; // 00 - 99
var match3 = /\d{3}/; // 000 - 999
var match4 = /\d{4}/; // 0000 - 9999
var match1to2 = /\d\d?/; // 0 - 99
var matchShortOffset = /[+-]\d\d:?\d\d/; // +00:00 -00:00 +0000 or -0000
var matchSigned = /[+-]?\d+/; // -inf - inf
var matchTimestamp = /[+-]?\d+(\.\d{1,3})?/; // 123456789 123456789.123
var matchWord = /[0-9]{0,256}['a-z\u00A0-\u05FF\u0700-\uD7FF\uF900-\uFDCF\uFDF0-\uFF07\uFF10-\uFFEF]{1,256}|[\u0600-\u06FF\/]{1,256}(\s*?[\u0600-\u06FF]{1,256}){1,2}/i; // Word
var YEAR = 'year';
var MONTH = 'month';
var DAY = 'day';
var HOUR = 'hour';
var MINUTE = 'minute';
var SECOND = 'second';
var MILLISECOND = 'millisecond';
var parseFlags = {};
var addParseFlag = function (token, regex, callback) {
    var tokens = Array.isArray(token) ? token : [token];
    var func;
    if (typeof callback === 'string') {
        func = function (input) {
            var _a;
            var value = parseInt(input, 10);
            return _a = {}, _a[callback] = value, _a;
        };
    }
    else {
        func = callback;
    }
    tokens.forEach(function (key) {
        parseFlags[key] = [regex, func];
    });
};
var matchWordCallback = function (localeKey, key) {
    return function (input, locale) {
        var _a;
        var array = locale[localeKey];
        if (!Array.isArray(array)) {
            throw new Error("Locale[" + localeKey + "] need an array");
        }
        var index = array.indexOf(input);
        if (index < 0) {
            throw new Error('Invalid Word');
        }
        return _a = {}, _a[key] = index, _a;
    };
};
addParseFlag('YY', match2, function (input) {
    var _a;
    var year = new Date().getFullYear();
    var cent = Math.floor(year / 100);
    var value = parseInt(input, 10);
    value = (value > 68 ? cent - 1 : cent) * 100 + value;
    return _a = {}, _a[YEAR] = value, _a;
});
addParseFlag('YYYY', match4, YEAR);
addParseFlag('M', match1to2, function (input) {
    var _a;
    return (_a = {}, _a[MONTH] = parseInt(input, 10) - 1, _a);
});
addParseFlag('MM', match2, function (input) {
    var _a;
    return (_a = {}, _a[MONTH] = parseInt(input, 10) - 1, _a);
});
addParseFlag('MMM', matchWord, matchWordCallback('monthsShort', MONTH));
addParseFlag('MMMM', matchWord, matchWordCallback('months', MONTH));
addParseFlag('D', match1to2, DAY);
addParseFlag('DD', match2, DAY);
addParseFlag(['H', 'h'], match1to2, HOUR);
addParseFlag(['HH', 'hh'], match2, HOUR);
addParseFlag('m', match1to2, MINUTE);
addParseFlag('mm', match2, MINUTE);
addParseFlag('s', match1to2, SECOND);
addParseFlag('ss', match2, SECOND);
addParseFlag('S', match1, function (input) {
    var _a;
    return _a = {},
        _a[MILLISECOND] = parseInt(input, 10) * 100,
        _a;
});
addParseFlag('SS', match2, function (input) {
    var _a;
    return _a = {},
        _a[MILLISECOND] = parseInt(input, 10) * 10,
        _a;
});
addParseFlag('SSS', match3, MILLISECOND);
function matchMeridiem(locale) {
    return locale.meridiemParse || /[ap]\.?m?\.?/i;
}
function defaultIsPM(input) {
    return ("" + input).toLowerCase().charAt(0) === 'p';
}
addParseFlag(['A', 'a'], matchMeridiem, function (input, locale) {
    var isPM = typeof locale.isPM === 'function' ? locale.isPM(input) : defaultIsPM(input);
    return { isPM: isPM };
});
function offsetFromString(str) {
    var _a = str.match(/([+-]|\d\d)/g) || ['-', '0', '0'], symbol = _a[0], hour = _a[1], minute = _a[2];
    var minutes = parseInt(hour, 10) * 60 + parseInt(minute, 10);
    if (minutes === 0) {
        return 0;
    }
    return symbol === '+' ? -minutes : +minutes;
}
addParseFlag(['Z', 'ZZ'], matchShortOffset, function (input) {
    return { offset: offsetFromString(input) };
});
addParseFlag('x', matchSigned, function (input) {
    return { date: new Date(parseInt(input, 10)) };
});
addParseFlag('X', matchTimestamp, function (input) {
    return { date: new Date(parseFloat(input) * 1000) };
});
addParseFlag('d', match1, 'weekday');
addParseFlag('dd', matchWord, matchWordCallback('weekdaysMin', 'weekday'));
addParseFlag('ddd', matchWord, matchWordCallback('weekdaysShort', 'weekday'));
addParseFlag('dddd', matchWord, matchWordCallback('weekdays', 'weekday'));
addParseFlag('w', match1to2, 'week');
addParseFlag('ww', match2, 'week');
function to24hour(hour, isPM) {
    if (hour !== undefined && isPM !== undefined) {
        if (isPM) {
            if (hour < 12) {
                return hour + 12;
            }
        }
        else if (hour === 12) {
            return 0;
        }
    }
    return hour;
}
function getFullInputArray(input, backupDate) {
    if (backupDate === void 0) { backupDate = new Date(); }
    var result = [0, 0, 1, 0, 0, 0, 0];
    var backupArr = [
        backupDate.getFullYear(),
        backupDate.getMonth(),
        backupDate.getDate(),
        backupDate.getHours(),
        backupDate.getMinutes(),
        backupDate.getSeconds(),
        backupDate.getMilliseconds(),
    ];
    var useBackup = true;
    for (var i = 0; i < 7; i++) {
        if (input[i] === undefined) {
            result[i] = useBackup ? backupArr[i] : result[i];
        }
        else {
            result[i] = input[i];
            useBackup = false;
        }
    }
    return result;
}
function createUTCDate() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var date;
    var y = args[0];
    if (y < 100 && y >= 0) {
        args[0] += 400;
        date = new Date(Date.UTC.apply(Date, args));
        // eslint-disable-next-line no-restricted-globals
        if (isFinite(date.getUTCFullYear())) {
            date.setUTCFullYear(y);
        }
    }
    else {
        date = new Date(Date.UTC.apply(Date, args));
    }
    return date;
}
function makeParser(dateString, format, locale) {
    var tokens = format.match(formattingTokens);
    if (!tokens) {
        throw new Error();
    }
    var length = tokens.length;
    var mark = {};
    for (var i = 0; i < length; i += 1) {
        var token = tokens[i];
        var parseTo = parseFlags[token];
        if (!parseTo) {
            var word = token.replace(/^\[|\]$/g, '');
            if (dateString.indexOf(word) === 0) {
                dateString = dateString.substr(word.length);
            }
            else {
                throw new Error('not match');
            }
        }
        else {
            var regex = typeof parseTo[0] === 'function' ? parseTo[0](locale) : parseTo[0];
            var parser = parseTo[1];
            var value = (regex.exec(dateString) || [])[0];
            var obj = parser(value, locale);
            mark = __assign(__assign({}, mark), obj);
            dateString = dateString.replace(value, '');
        }
    }
    return mark;
}
export default function parse(str, format, options) {
    if (options === void 0) { options = {}; }
    try {
        var _a = options.locale, locale = _a === void 0 ? defaultLocale : _a, _b = options.backupDate, backupDate = _b === void 0 ? new Date() : _b;
        var parseResult = makeParser(str, format, locale);
        var year = parseResult.year, month = parseResult.month, day = parseResult.day, hour = parseResult.hour, minute = parseResult.minute, second = parseResult.second, millisecond = parseResult.millisecond, isPM = parseResult.isPM, date = parseResult.date, offset = parseResult.offset, weekday = parseResult.weekday, week = parseResult.week;
        if (date) {
            return date;
        }
        var inputArray = [year, month, day, hour, minute, second, millisecond];
        inputArray[3] = to24hour(inputArray[3], isPM);
        // check week
        if (week !== undefined && month === undefined && day === undefined) {
            // new Date(year, 3) make sure in current year
            var firstDate = startOfWeekYear(year === undefined ? backupDate : new Date(year, 3), {
                firstDayOfWeek: locale.firstDayOfWeek,
                firstWeekContainsDate: locale.firstWeekContainsDate,
            });
            return new Date(firstDate.getTime() + (week - 1) * 7 * 24 * 3600 * 1000);
        }
        var utcDate = createUTCDate.apply(void 0, getFullInputArray(inputArray, backupDate));
        var offsetMilliseconds = (offset === undefined ? utcDate.getTimezoneOffset() : offset) * 60 * 1000;
        var parsedDate = new Date(utcDate.getTime() + offsetMilliseconds);
        // check weekday
        if (weekday !== undefined && parsedDate.getDay() !== weekday) {
            return new Date(NaN);
        }
        return parsedDate;
    }
    catch (e) {
        return new Date(NaN);
    }
}
