"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function isDate(value) {
    return value instanceof Date || Object.prototype.toString.call(value) === '[object Date]';
}
exports.isDate = isDate;
function toDate(value) {
    if (isDate(value)) {
        return new Date(value.getTime());
    }
    if (value == null) {
        return new Date(NaN);
    }
    return new Date(value);
}
exports.toDate = toDate;
function isValidDate(value) {
    return isDate(value) && !isNaN(value.getTime());
}
exports.isValidDate = isValidDate;
function startOfWeek(value, firstDayOfWeek) {
    if (firstDayOfWeek === void 0) { firstDayOfWeek = 0; }
    if (!(firstDayOfWeek >= 0 && firstDayOfWeek <= 6)) {
        throw new RangeError('weekStartsOn must be between 0 and 6');
    }
    var date = toDate(value);
    var day = date.getDay();
    var diff = (day + 7 - firstDayOfWeek) % 7;
    date.setDate(date.getDate() - diff);
    date.setHours(0, 0, 0, 0);
    return date;
}
exports.startOfWeek = startOfWeek;
function startOfWeekYear(value, _a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.firstDayOfWeek, firstDayOfWeek = _c === void 0 ? 0 : _c, _d = _b.firstWeekContainsDate, firstWeekContainsDate = _d === void 0 ? 1 : _d;
    if (!(firstWeekContainsDate >= 1 && firstWeekContainsDate <= 7)) {
        throw new RangeError('firstWeekContainsDate must be between 1 and 7');
    }
    var date = toDate(value);
    var year = date.getFullYear();
    var firstDateOfFirstWeek = new Date(0);
    for (var i = year + 1; i >= year - 1; i--) {
        firstDateOfFirstWeek.setFullYear(i, 0, firstWeekContainsDate);
        firstDateOfFirstWeek = startOfWeek(firstDateOfFirstWeek, firstDayOfWeek);
        if (date.getTime() >= firstDateOfFirstWeek.getTime()) {
            break;
        }
    }
    return firstDateOfFirstWeek;
}
exports.startOfWeekYear = startOfWeekYear;
function getWeek(value, _a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.firstDayOfWeek, firstDayOfWeek = _c === void 0 ? 0 : _c, _d = _b.firstWeekContainsDate, firstWeekContainsDate = _d === void 0 ? 1 : _d;
    var date = toDate(value);
    var firstDateOfThisWeek = startOfWeek(date, firstDayOfWeek);
    var firstDateOfFirstWeek = startOfWeekYear(date, { firstDayOfWeek: firstDayOfWeek, firstWeekContainsDate: firstWeekContainsDate });
    var diff = firstDateOfThisWeek.getTime() - firstDateOfFirstWeek.getTime();
    return Math.round(diff / (7 * 24 * 3600 * 1000)) + 1;
}
exports.getWeek = getWeek;
